# HTTP Stubber

- [HTTP Stubber](#http-stubber)
  - [Stub a Request](#stub-a-request)
  - [Stub Matchers](#stub-matchers)
    - [Echo Matcher](#echo-matcher)
    - [Dynamic Matcher](#dynamic-matcher)
    - [URI Matcher](#uri-matcher)
    - [JSON Matcher](#json-matcher)
    - [Body Matcher](#body-matcher)
    - [URL Matcher](#url-matcher)
    - [Custom Matcher](#custom-matcher)
  - [Add Ignore Rule](#add-ignore-rule)
  - [Unhandled Rules](#unhandled-rules)

RealHTTP offer an elegant support for stubbing HTTP requests in Swift, allowing you to stub any HTTP/HTTPS using `URLSession`. That includes any request made from libraries such as RealHTTP but also with Alamofire.


To activate and configure the stubber you need to use the `HTTPStubber.shared` instance:

```swift
HTTPStubber.shared.enable() // enable the stubber
```

Once you have done you can disable it:

```swift
HTTPStubber.shared.disable()
```

## Stub a Request

To stub a request, first you need to create a `HTTPStubRequest` and `HTTPStubResponse`.  
You then register this stub with RealHTTP and tell it to intercept network requests by calling the `enable()` method.

An `HTTPStubRequest` describe what kind of triggers should be matched for activating the stub request; a single trigger is called matcher which is an entity conforms to `HTTPStubMatcherProtocol`.  
Only when all `matchers` of a stub request are validated the stub is used to mock the request; in this case the associated response are used.  
You can specify a different `HTTPStubResponse` for each HTTP Method (ie. a different response for GET and POST on the same trigger).

```swift
var stub = try HTTPStubRequest()
           .match(urlRegex: "http://www.google.com/+") // stub any url for google domain
           .stub(for: .post, delay: 5, json: jsonData) // stub the post with json data and delay
           .stub(for: .put, error: internalError) // stub error for put

// Add to the stubber
HTTPStubber.shared.add(stub: stub)
```

The following example use the URL Regex Matcher to match any url of the google domain.  
The stub return
- a `jsonData` for `POST` requests with a delay of the response of 5 seconds
- an `internalError` for any `PUT` request

an `HTTPStubResponse` is an object associated to a `HTTPStubRequest` for a particular HTTP Method. Each object can be configured to return custom data such as:

- `s
- tatusCode`: the HTTP Status Code
- `contentType`: the content-type header of the response
- `failError`: when specified a non nil error the `Error` is stubbed as response
- `body`: the body of the response
- `headers`: headers to return with the response
- `cachePolicy`: cache policy to trigger for resposne
- `responseDelay`: delay interval in seconds before responding to a triggere request. Very useful for testing.

While you can use any of the shortcuts in `HTTPStubRequest`'s `stub(...)` functions to set coincise `HTTPStubResponse` you can also provide a complete object to specify the properties above:

```swift
var stub = HTTPStubRequest()
           .stub(for: .post, { response in
                // Builder function allows you to specifiy any property of the
                // HTTPStubResponse object for this http method.
                response.responseDelay = 5
                response.headers = HTTPHeaders([
                    .contentType: HTTPContentType.bmp.rawValue,
                    .contentLength: String(fileSize,
                ])
                response.body = fileContent
            })
```

## Stub Matchers

There are different stub matchers you can chain for a single stub request.

> NOTE: `matchers` are evaluated with AND operator, not OR. So all matchers must be verified in order to trigger the relative request.

### Echo Matcher

A special stubber is `HTTPEchoResponse` which basically allows to respond with the same request received.  
This is particularly useful when you need to check and validate your requests.

To enable it:

```swift
let echoStub = HTTPStubRequest().match(urlRegex: "*").stubEcho()
HTTPStubber.shared.add(stub: echoStub)
HTTPStubber.shared.enable()
```

### Dynamic Matcher

Sometimes you may want to customize the `HTTPStubResponse` based upon certains parameters of the received `URLRequest`/`HTTPStubRequest`.  
RealHTTP offer a way to return the appropriate stub response based upon received request using the `stub(method:responseProvider:)` function which creates the `HTTPDynamicStubResponse` instance.

The following example capture the `login` call and allows to personalize the `HTTPStubResponse` to return:

```swift

let stubLogin = HTTPStubRequest().match(urlRegex: "/login").stub(for: .post, responseProvider: { urlRequest, matchedStubRequest in
    let response = HTTPStubResponse()
            
    guard let dict = try! JSONSerialization.jsonObject(with: urlRequest.body!) as? NSDictionary,
          let username = dict["user"] as? String else {
      response.statusCode = .unauthorized
      return response
    }
            
    if username == "mark" {
       // configure the response for mark, our special admin!
    } else {
      // ... return another response for anyone else!
    }
            
    return response
})
```

### URI Matcher

RealHTTP allows you to define an URI matcher (`HTTPStubURITemplateMatcher`) which is based upon the [URI Template RFC](https://tools.ietf.org/html/rfc6570) - it uses the [URITemplate project by Kyle Fuller](https://github.com/kylef/URITemplate.swift).

```swift
var stub = HTTPStubRequest()
           .match(URI: "https://github.com/kylef/{repository}")
           .stub(...)
```

The uri function takes a URL or path which can have a URI Template.
Such as the following:

- https://github.com/kylef/WebLinking.swift
- https://github.com/kylef/{repository}
- /kylef/{repository}
- /kylef/URITemplate.swift

### JSON Matcher

Say you're POSTing a JSON to your server, you could make your stub match a particular value like this:

```swift
var stub = HTTPStubRequest()
           .match(object: User(userID: 34, fullName: "Mark"))
```

It will match the JSON representation of an `User` struct with `userID=34` and `fullName=Mark` without representing the raw JSON but just passing the `Codable` struct!  
It uses the `HTTPStubJSONMatcher` matcher.

### Body Matcher

Using the `HTTPStubBodyMatcher` you can match the body of a request which should be equal to a particular value in order to trigger the relative stub request:

```swift
let bodyToCheck: Data = ...
var stub = HTTPStubRequest()
           .match(body: bodyToCheck)
           .stub(...)
```

### URL Matcher

The URL matcher `HTTPStubURLMatcher` is a simple version of the regular expression matcher which check the equality of an URL by including or ignoring the query parameters:

```swift
var stub = HTTPStubRequest()
           .match(URL: "http://myws.com/login", options: .ignoreQueryParameters)
```

It will match URLs by ignoring any query parameter like `http://myws.com/login?username=xxxx` or `http://myws.com/login?username=yyyy&lock=1`.

### Custom Matcher

You can create your custom matcher and add it to the `matchers` of a stub request. It must be conform to the `HTTPStubMatcherProtocol` protocol:

```swift
public protocol HTTPStubMatcherProtocol {
    func matches(request: URLRequest, for source: HTTPMatcherSource) -> Bool
}
```

When you are ready use `add()` to add the rule:

```swift
struct MyCustomMatcher: HTTPStubMatcherProtocol { ... } // implement your logic

stub.match(MyCustomMatcher()) // add to the matcher of a request
```

Sometimes you may not want to create a custom object to validate your matcher (`HTTPStubCustomMatcher`):

```swift
var stub = HTTPStubRequest()
           .match({ req, source in
              req.headers[.userAgent] == "customAgent"
           })
           .stub(...)
```

The following stub is triggered when headers of the request is `customAgent`.

## Add Ignore Rule

You can add ignoring rule to the `HTTPStubber.shared` in order to ignore and pass through some URLs.  
An ignore rule is an object of type `HTTPStubIgnoreRule` which contains a list of `matchers`: when all matchers are verified the stub is valid and the request is ignored.  It works like stub request for response but for ignores.

```swift
HTTPStubber.shared.add(ignoreURL: "http://www.apple.com", options: [.ignorePath, .ignoreQueryParameters])
```

You can use all the matchers described above even for rules.

## Unhandled Rules

You can choose how RealHTTP must deal with stub not found. You have two options:

 - `optout`: only URLs registered wich matches the matchers are ignored for mocking.
            - Registered mocked URL: mocked.
             - Registered ignored URL: ignored by the stubber, default process is applied as if the stubber is disabled.
             - Any other URL: Raises an error.
 - `optin`: Only registered mocked URLs are mocked, all others pass through.
             - Registered mocked URL: mocked.
             - Any other URL: ignored by the stubber, default process is applied as if the stubber is disabled.

To set the behaviour:

```swift
HTTPStubber.shared.unhandledMode = .optin // enable optin mode
```
